from datasette.app import Datasette
from datasette_alerts import (
    Notifier,
    InternalDB,
    NewAlertRouteParameters,
    NewSubscription,
)
from wtforms import Form, StringField
import pytest
import pytest_asyncio
import sqlite3
import asyncio
import json


class MockNotifier(Notifier):
    """A mock notifier for testing."""

    @property
    def slug(self):
        return "mock-notifier"

    @property
    def name(self):
        return "Mock Notifier"

    description = "A test notifier"
    icon = "🔔"

    def __init__(self):
        self.sent_messages = []

    async def send(self, alert_id, new_ids, meta):
        """Record sent messages for testing."""
        self.sent_messages.append(
            {"alert_id": alert_id, "new_ids": new_ids, "meta": meta}
        )

    async def get_config_form(self):
        """Return a simple config form."""

        class ConfigForm(Form):
            url = StringField("URL")

        return ConfigForm


@pytest_asyncio.fixture
async def datasette(tmpdir):
    """Create a test Datasette instance with a sample database."""
    data = str(tmpdir / "data.db")
    db = sqlite3.connect(data)
    with db:
        db.execute(
            """
            create table events (
                id integer primary key,
                title text,
                created_at timestamp default current_timestamp
            )
        """
        )
        db.execute(
            """
            insert into events (title, created_at)
            values ('Event 1', '2024-01-01 10:00:00')
        """
        )
        db.execute(
            """
            insert into events (title, created_at)
            values ('Event 2', '2024-01-01 11:00:00')
        """
        )

    datasette = Datasette([data])
    datasette._test_db = db
    await datasette.invoke_startup()
    return datasette


@pytest.mark.asyncio
async def test_plugin_loads_and_creates_tables(datasette):
    """Test that the plugin loads and creates internal tables."""
    internal_db = datasette.get_internal_database()

    # Check that internal tables exist
    tables = await internal_db.execute(
        "SELECT name FROM sqlite_master WHERE type='table'"
    )
    table_names = [row[0] for row in tables.rows]

    assert "datasette_alerts_alerts" in table_names
    assert "datasette_alerts_subscriptions" in table_names
    assert "datasette_alerts_alert_logs" in table_names


@pytest.mark.asyncio
async def test_internal_db_new_alert(datasette):
    """Test creating a new alert."""
    internal_db = InternalDB(datasette.get_internal_database())

    params = NewAlertRouteParameters(
        database_name="data",
        table_name="events",
        id_columns=["id"],
        timestamp_column="created_at",
        frequency="+1 hour",
        subscriptions=[
            NewSubscription(
                notifier_slug="test-notifier", meta={"url": "https://example.com"}
            )
        ],
    )

    alert_id = await internal_db.new_alert(params, "2024-01-01 11:00:00")

    # Verify alert was created
    assert alert_id is not None

    # Check alert details
    db = datasette.get_internal_database()
    result = await db.execute(
        "SELECT * FROM datasette_alerts_alerts WHERE id = ?", [alert_id]
    )
    alert = dict(result.first())

    assert alert["database_name"] == "data"
    assert alert["table_name"] == "events"
    assert json.loads(alert["id_columns"]) == ["id"]
    assert alert["timestamp_column"] == "created_at"
    assert alert["frequency"] == "+1 hour"

    # Check subscription was created
    subscriptions = await db.execute(
        "SELECT * FROM datasette_alerts_subscriptions WHERE alert_id = ?", [alert_id]
    )
    sub = dict(subscriptions.first())
    assert sub["notifier"] == "test-notifier"
    assert json.loads(sub["meta"]) == {"url": "https://example.com"}

    # Check initial log entry was created
    logs = await db.execute(
        "SELECT * FROM datasette_alerts_alert_logs WHERE alert_id = ?", [alert_id]
    )
    log = dict(logs.first())
    assert json.loads(log["new_ids"]) == []
    assert log["cursor"] == "2024-01-01 11:00:00"


@pytest.mark.asyncio
async def test_internal_db_alert_subscriptions(datasette):
    """Test fetching alert subscriptions."""
    internal_db = InternalDB(datasette.get_internal_database())

    params = NewAlertRouteParameters(
        database_name="data",
        table_name="events",
        id_columns=["id"],
        timestamp_column="created_at",
        frequency="+1 hour",
        subscriptions=[
            NewSubscription(notifier_slug="notifier1", meta={"key1": "value1"}),
            NewSubscription(notifier_slug="notifier2", meta={"key2": "value2"}),
        ],
    )

    alert_id = await internal_db.new_alert(params, "2024-01-01 00:00:00")

    # Fetch subscriptions
    subscriptions = await internal_db.alert_subscriptions(alert_id)

    assert len(subscriptions) == 2
    assert subscriptions[0]["notifier"] == "notifier1"
    assert json.loads(subscriptions[0]["meta"]) == {"key1": "value1"}
    assert subscriptions[1]["notifier"] == "notifier2"
    assert json.loads(subscriptions[1]["meta"]) == {"key2": "value2"}


@pytest.mark.asyncio
async def test_internal_db_add_log(datasette):
    """Test adding a log entry."""
    internal_db = InternalDB(datasette.get_internal_database())

    params = NewAlertRouteParameters(
        database_name="data",
        table_name="events",
        id_columns=["id"],
        timestamp_column="created_at",
        frequency="+1 hour",
        subscriptions=[],
    )

    alert_id = await internal_db.new_alert(params, "2024-01-01 00:00:00")

    # Add a log entry
    await internal_db.add_log(alert_id, ["1", "2", "3"], "2024-01-01 12:00:00")

    # Verify log entry
    db = datasette.get_internal_database()
    logs = await db.execute(
        "SELECT * FROM datasette_alerts_alert_logs WHERE alert_id = ? ORDER BY logged_at ASC",
        [alert_id],
    )
    rows = list(logs.rows)

    # Should have 2 logs: initial + new one
    assert len(rows) == 2

    # First log is the initial empty one
    initial_log = dict(rows[0])
    assert json.loads(initial_log["new_ids"]) == []
    assert initial_log["cursor"] == "2024-01-01 00:00:00"

    # Second log is the new one we added
    latest_log = dict(rows[1])
    assert json.loads(latest_log["new_ids"]) == ["1", "2", "3"]
    assert latest_log["cursor"] == "2024-01-01 12:00:00"


@pytest.mark.asyncio
async def test_internal_db_schedule_next(datasette):
    """Test scheduling the next alert run."""
    internal_db = InternalDB(datasette.get_internal_database())

    params = NewAlertRouteParameters(
        database_name="data",
        table_name="events",
        id_columns=["id"],
        timestamp_column="created_at",
        frequency="+1 hour",
        subscriptions=[],
    )

    alert_id = await internal_db.new_alert(params, "2024-01-01 00:00:00")

    # Simulate starting the job
    db = datasette.get_internal_database()
    await db.execute_write(
        "UPDATE datasette_alerts_alerts SET current_schedule_started_at = datetime('now') WHERE id = ?",
        [alert_id],
    )

    # Schedule next run
    await internal_db.schedule_next(alert_id)

    # Verify next_deadline was updated and current_schedule_started_at was cleared
    result = await db.execute(
        "SELECT next_deadline, current_schedule_started_at FROM datasette_alerts_alerts WHERE id = ?",
        [alert_id],
    )
    row = dict(result.first())

    assert row["next_deadline"] is not None
    assert row["current_schedule_started_at"] is None


@pytest.mark.asyncio
async def test_internal_db_start_ready_jobs(datasette):
    """Test finding and starting ready jobs."""
    internal_db = InternalDB(datasette.get_internal_database())

    # Create an alert with a deadline in the past
    db = datasette.get_internal_database()

    params = NewAlertRouteParameters(
        database_name="data",
        table_name="events",
        id_columns=["id"],
        timestamp_column="created_at",
        frequency="+1 hour",
        subscriptions=[],
    )

    alert_id = await internal_db.new_alert(params, "2024-01-01 00:00:00")

    # Set the deadline to the past
    await db.execute_write(
        "UPDATE datasette_alerts_alerts SET next_deadline = datetime('now', '-1 hour') WHERE id = ?",
        [alert_id],
    )

    # Get ready jobs
    ready_jobs = await internal_db.start_ready_jobs()

    assert len(ready_jobs) == 1
    job = ready_jobs[0]
    assert job.alert_id == alert_id
    assert job.database_name == "data"
    assert job.table_name == "events"
    assert job.id_columns == ["id"]
    assert job.timestamp_column == "created_at"
    assert job.cursor == "2024-01-01 00:00:00"

    # Verify current_schedule_started_at was set
    result = await db.execute(
        "SELECT current_schedule_started_at FROM datasette_alerts_alerts WHERE id = ?",
        [alert_id],
    )
    row = dict(result.first())
    assert row["current_schedule_started_at"] is not None


@pytest.mark.asyncio
async def test_api_new_alert_endpoint(datasette):
    """Test the API endpoint for creating alerts."""
    payload = {
        "database_name": "data",
        "table_name": "events",
        "id_columns": ["id"],
        "timestamp_column": "created_at",
        "frequency": "+1 hour",
        "subscriptions": [
            {"notifier_slug": "test-notifier", "meta": {"url": "https://example.com"}}
        ],
    }

    response = await datasette.client.post(
        "/-/datasette-alerts/api/new-alert", json=payload
    )

    assert response.status_code == 200
    data = response.json()
    assert data["ok"] is True
    assert "alert_id" in data["data"]

    # Verify alert was created
    alert_id = data["data"]["alert_id"]
    db = datasette.get_internal_database()
    result = await db.execute(
        "SELECT * FROM datasette_alerts_alerts WHERE id = ?", [alert_id]
    )
    alert = dict(result.first())
    assert alert["table_name"] == "events"


@pytest.mark.asyncio
async def test_api_new_alert_invalid_database(datasette):
    """Test API endpoint with invalid database name."""
    payload = {
        "database_name": "nonexistent",
        "table_name": "events",
        "id_columns": ["id"],
        "timestamp_column": "created_at",
        "frequency": "+1 hour",
        "subscriptions": [],
    }

    response = await datasette.client.post(
        "/-/datasette-alerts/api/new-alert", json=payload
    )

    assert response.status_code == 404
    data = response.json()
    assert data["ok"] is False
    assert "not found" in data["error"]


@pytest.mark.asyncio
async def test_api_new_alert_invalid_payload(datasette):
    """Test API endpoint with invalid payload."""
    response = await datasette.client.post(
        "/-/datasette-alerts/api/new-alert", json={"invalid": "payload"}
    )

    assert response.status_code == 400
    data = response.json()
    assert data["ok"] is False


@pytest.mark.asyncio
async def test_api_new_alert_wrong_method(datasette):
    """Test API endpoint with GET instead of POST."""
    response = await datasette.client.get("/-/datasette-alerts/api/new-alert")

    assert response.status_code == 405


@pytest.mark.asyncio
async def test_table_action_link_with_permission(datasette):
    """Test that the table action link appears when user has permission."""
    # Set root_enabled to allow root permissions
    datasette.root_enabled = True

    # Create a signed cookie for root user
    cookies = {"ds_actor": datasette.sign({"a": {"id": "root"}}, "actor")}

    # Visit the table page
    response = await datasette.client.get("/data/events", cookies=cookies)
    assert response.status_code == 200

    # Check that the alert configuration link is present
    assert "Configure new alert" in response.text
    assert (
        "/-/datasette-alerts/new-alert?db_name=data&amp;table_name=events"
        in response.text
    )


@pytest.mark.asyncio
async def test_table_action_link_without_permission(datasette):
    """Test that the table action link does not appear without permission."""
    # Visit the table page without authentication
    response = await datasette.client.get("/data/events")
    assert response.status_code == 200

    # Check that the alert configuration link is NOT present
    assert "Configure new alert" not in response.text
    assert "datasette-alerts/new-alert" not in response.text



# https://github.com/binwiederhier/ntfy/blob/main/web/src/img/ntfy-outline.svg

from datasette import hookimpl
from datasette_alerts import Notifier
import httpx
from wtforms import Form, StringField


@hookimpl
def datasette_alerts_register_notifiers(datasette):
    return [Ntfy()]


class Ntfy(Notifier):
    slug = "ntfy"
    name = "Ntfy"
    description = "Send alerts to ntfy.sh"
    icon = '<svg width="22px" height="22px" viewBox="0 0 50 50" xmlns="http://www.w3.org/2000/svg"><g style="display:inline"><path style="color:#000;fill:#777;fill-opacity:1;stroke:none;stroke-width:.754022;-inkscape-stroke:none" d="M59.292 93.677c-3.58 0-6.647 2.817-6.647 6.398v.003l.035 27.867-.9 6.635 12.227-3.248H94.4c3.58 0 6.646-2.82 6.646-6.402v-24.855c0-3.58-3.065-6.396-6.643-6.398H94.4zm0 4.516h35.112c1.257.002 2.127.917 2.127 1.882v24.855c0 .966-.871 1.882-2.13 1.882H63.344l-6.211 1.877.063-.366-.034-28.248c0-.966.87-1.882 2.13-1.882z" transform="translate(-51.452 -87.327)"/><g style="font-size:8.48274px;font-family:sans-serif;letter-spacing:0;word-spacing:0;fill:#777;fill-opacity:1;stroke:none;stroke-width:.525121"><path style="color:#000;-inkscape-font-specification:\'JetBrains Mono, Bold\';fill:#777;fill-opacity:1;stroke:none;-inkscape-stroke:none" d="M62.57 116.77v-1.312l3.28-1.459q.159-.068.306-.102.158-.045.283-.068l.271-.022v-.09q-.136-.012-.271-.046-.125-.023-.283-.057-.147-.045-.306-.113l-3.28-1.459v-1.323l5.068 2.319v1.413z" transform="matrix(2.14521 0 0 2.55031 -122.7 -265.715)"/><path style="color:#000;-inkscape-font-specification:\'JetBrains Mono, Bold\';fill:#777;fill-opacity:1;stroke:none;-inkscape-stroke:none" d="M62.309 110.31v1.903l3.437 1.53.022.007-.022.008-3.437 1.53v1.892l.37-.17 5.221-2.39v-1.75zm.525.817 4.541 2.08v1.076l-4.541 2.078v-.732l3.12-1.389.003-.002a1.56 1.56 0 0 1 .258-.086h.006l.008-.002c.094-.027.176-.047.246-.06l.498-.041v-.574l-.24-.02a1.411 1.411 0 0 1-.231-.04l-.008-.001-.008-.002a9.077 9.077 0 0 1-.263-.053 2.781 2.781 0 0 1-.266-.097l-.004-.002-3.119-1.39z" transform="matrix(2.14521 0 0 2.55031 -122.7 -265.715)"/></g><g style="font-size:8.48274px;font-family:sans-serif;letter-spacing:0;word-spacing:0;fill:#777;fill-opacity:1;stroke:none;stroke-width:.525121"><path style="color:#000;-inkscape-font-specification:\'JetBrains Mono, Bold\';fill:#777;fill-opacity:1;stroke:none;-inkscape-stroke:none" d="M69.171 117.754h5.43v1.278h-5.43Z" transform="matrix(2.13886 0 0 2.45586 -121.197 -258.267)"/><path style="color:#000;-inkscape-font-specification:\'JetBrains Mono, Bold\';fill:#777;fill-opacity:1;stroke:none;-inkscape-stroke:none" d="M68.908 117.492v1.802h5.955v-1.802zm.526.524h4.904v.754h-4.904z" transform="matrix(2.13886 0 0 2.45586 -121.197 -258.267)"/></g></g></svg>'

    def __init__(self):
        pass

    async def get_config_form(self):
        class ConfigForm(Form):
            base_url = StringField(
                "Base URL",
                description="...",
                default="https://ntfy.sh",
            )
            topic = StringField(
                "Topic",
                description="...",
            )

        return ConfigForm

    async def send(self, alert_id, new_ids, config: dict):
        title = f"{len(new_ids)} new items in {'TODO'}"
        message = "\n".join([f"- `{id}` " for id in new_ids])
        link = "https://datasette.io"
        # https://docs.ntfy.sh/publish/#publish-as-json
        print(config)
        send_ntfy(
            config["base_url"],
            config["topic"],
            title,
            message,
            link,
        )


def send_ntfy(url, topic, title, message, link):
    json = {
        "topic": topic,
        "title": title,
        "message": message,
        # "click": {"link": link},
        "markdown": True,
    }
    httpx.post(
        url,
        json=json,
    )



# /// script
# requires-python = ">=3.13"
# dependencies = [
#     "click",
#     "httpx",
# ]
# ///

import click
import sqlite3
import httpx
import json
from datetime import date

SCHEMA = """
CREATE TABLE IF NOT EXISTS federal_register_documents(
  document_number text PRIMARY KEY,
  title text,
  abstract text,
  document_type text,
  agencies json,
  excerpts text,
  html_url text,
  pdf_url text,
  public_inspection_pdf_url text
);

CREATE TABLE IF NOT EXISTS federal_register_agencies(
  id integer PRIMARY KEY,
  parent_id integer REFERENCES federal_register_agencies(id),
  slug text,
  name text,
  short_name text,
  description text,
  url text,
  child_ids json
);
"""


@click.group()
def cli():
    """Federal Register data processing tools."""
    pass


def backfill_agencies(db):
    """Backfill Federal Register agencies into the database."""
    response = httpx.get("https://www.federalregister.gov/api/v1/agencies").json()

    with db:
        for agency in response:
            db.execute(
                "INSERT OR IGNORE INTO federal_register_agencies (id, parent_id, slug, name, short_name, description, url, child_ids) VALUES (?, ?, ?, ?, ?, ?, ?, ?)",
                (
                    agency["id"],
                    agency.get("parent_id"),
                    agency["slug"],
                    agency["name"],
                    agency.get("short_name", ""),
                    agency.get("description", ""),
                    agency.get("url", ""),
                    json.dumps(agency.get("child_ids", [])),
                ),
            )


@cli.command()
@click.argument("output")
@click.option(
    "--until",
    required=False,
    type=click.DateTime(formats=["%Y-%m-%d"]),
    help="End date in YYYY-MM-DD format",
    default=date.today(),
)
def sync(output, until):
    """Backfill Federal Register data into a database."""
    click.echo(f"Backfilling data to {output} until {until.strftime('%Y-%m-%d')}")
    db = sqlite3.connect(output)
    with db:
        db.executescript(SCHEMA)

    agency_id = db.execute(
        "select id from federal_register_agencies limit 1"
    ).fetchone()

    if not agency_id:
        backfill_agencies(db)

    response = httpx.get(
        f"https://www.federalregister.gov/api/v1/documents.json?per_page=1000&conditions[publication_date][lte]={until}"
    ).json()

    changes = 0
    with db:
        for result in response["results"]:
            document_number = result["document_number"]
            title = result["title"]
            abstract = result.get("abstract", "")
            document_type = result["type"]
            agencies = json.dumps([agency.get("id") for agency in result["agencies"]])
            excerpts = result.get("excerpts", "")
            html_url = result["html_url"]
            pdf_url = result.get("pdf_url", "")
            public_inspection_pdf_url = result.get("public_inspection_pdf_url", "")

            db.execute(
                "INSERT OR IGNORE INTO federal_register_documents (document_number, title, abstract, document_type, agencies, excerpts, html_url, pdf_url, public_inspection_pdf_url) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)",
                (
                    document_number,
                    title,
                    abstract,
                    document_type,
                    agencies,
                    excerpts,
                    html_url,
                    pdf_url,
                    public_inspection_pdf_url,
                ),
            )
            changes += db.execute("select changes()").fetchone()[0]

    click.echo("Loaded {} documents into the database.".format(changes))


def main() -> None:
    cli()


if __name__ == "__main__":
    main()



from notifypy import Notify


# https://icons.getbootstrap.com/icons/discord/

from datasette import hookimpl
from datasette_alerts import Notifier
import httpx
from wtforms import Form, StringField


@hookimpl
def datasette_alerts_register_notifiers(datasette):
    return [DesktopNotifier()]


class DesktopNotifier(Notifier):
    slug = "desktop"
    name = "Desktop"
    description = "Send alerts to your Desktop"
    icon = '<svg xmlns="http://www.w3.org/2000/svg" width="16" height="16" fill="currentColor" class="bi bi-terminal" viewBox="0 0 16 16"><path d="M6 9a.5.5 0 0 1 .5-.5h3a.5.5 0 0 1 0 1h-3A.5.5 0 0 1 6 9M3.854 4.146a.5.5 0 1 0-.708.708L4.793 6.5 3.146 8.146a.5.5 0 1 0 .708.708l2-2a.5.5 0 0 0 0-.708z"/><path d="M2 1a2 2 0 0 0-2 2v10a2 2 0 0 0 2 2h12a2 2 0 0 0 2-2V3a2 2 0 0 0-2-2zm12 1a1 1 0 0 1 1 1v10a1 1 0 0 1-1 1H2a1 1 0 0 1-1-1V3a1 1 0 0 1 1-1z"/></svg>'

    def __init__(self):
        pass

    async def get_config_form(self):
        class ConfigForm(Form):
            title = StringField()

        return ConfigForm

    async def send(self, alert_id, new_ids, config: dict):
        message = f"{len(new_ids)} new items in {'TODO'}"
        print(message)
        notification = Notify()
        notification.title = config["title"]
        notification.message = message
        notification.send()
